/*
 * \brief  Utility for accessing objects by name
 * \author Norman Feske
 * \date   2022-09-14
 */

/*
 * Copyright (C) 2022 Genode Labs GmbH
 *
 * This file is part of the Genode OS framework, which is distributed
 * under the terms of the GNU Affero General Public License version 3.
 */

#ifndef _INCLUDE__UTIL__DICTIONARY_H_
#define _INCLUDE__UTIL__DICTIONARY_H_

#include <util/meta.h>
#include <util/string.h>
#include <util/avl_tree.h>
#include <util/noncopyable.h>

namespace Genode { template <typename, typename> class Dictionary; }


template <typename T, typename NAME>
class Genode::Dictionary : Noncopyable
{
	private:

		Avl_tree<T> _tree { };

	public:

		class Element : private Avl_node<T>
		{
			public:

				NAME const name;

			private:

				Dictionary<T, NAME> &_registry;

				bool higher(T const *other) const { return name > other->name; }

				friend class Avl_tree<T>;
				friend class Avl_node<T>;
				friend class Dictionary<T, NAME>;

				static T *_matching_sub_tree(T &curr, NAME const &name)
				{
					typename Avl_node<T>::Side side = (curr.name > name);

					return curr.Avl_node<T>::child(side);
				}

			public:

				Element(Dictionary &registry, NAME const &name)
				:
					name(name), _registry(registry)
				{
					_registry._tree.insert(this);
				}

				~Element()
				{
					_registry._tree.remove(this);
				}
		};

		template <typename FN1, typename FN2>
		auto apply(NAME const &name, FN1 const &match_fn, FN2 const &no_match_fn)
		-> typename Trait::Functor<decltype(&FN1::operator())>::Return_type
		{
			T *curr_ptr = _tree.first();
			for (;;) {
				if (!curr_ptr)
					break;

				if (curr_ptr->name == name) {
					return match_fn(*curr_ptr);
				}

				curr_ptr = Element::_matching_sub_tree(*curr_ptr, name);
			}
			return no_match_fn();
		}

		template <typename FN1, typename FN2>
		auto apply(NAME const &name, FN1 const &match_fn, FN2 const &no_match_fn) const
		-> typename Trait::Functor<decltype(&FN1::operator())>::Return_type
		{
			auto const_match_fn = [&] (T const &e) { return match_fn(e); };
			auto non_const_this = const_cast<Dictionary *>(this);
			return non_const_this->apply(name, const_match_fn, no_match_fn);
		}

		template <typename ARG, typename FUNC>
		bool apply_any(FUNC const &fn)
		{
			T *curr_ptr = _tree.first();
			if (!curr_ptr)
				return false;

			fn(*curr_ptr);
			return true;
		}

		template <typename FN>
		void for_each(FN const &fn) const { _tree.for_each(fn); }
};

#endif /* _INCLUDE__UTIL__DICTIONARY_H_ */
